using System;
using System.Collections.Generic;
using System.Threading;

namespace Ilungasoft.Helper.ObjectPooling
{
    public class ObjectPool<ObjectType>
    {
        #region Private Members

        private bool canExhaust;
	    private SyncGuard sync;
        private bool created;
        private Dictionary<object, ObjectType> pool;
        private Stack<ObjectType> freeObjects;

        // Searches for a free item in the pool. If one is found, it is
        // returned, otherwise the method returns null to notify the
        // main Draw method (which would eventually grow the pool)
        private ObjectType InternalDraw()
        {
            sync.Lock();

            if (IsFreeObjectToMuch())
            {
                ShrinkPool();
            }

            // check for free objects
            if (freeObjects.Count > 0) // found?
            {
                ObjectType po = freeObjects.Pop();
                sync.Unlock();
                return po;
            }
            sync.Unlock();
            return default(ObjectType);
        }

        /// <summary>
        /// grows the pool as needed : creates objectCount number of objects
        /// </summary>
        /// <param name="objectCount">The object count.</param>
        /// <returns></returns>
        private bool Grow(int objectCount)
        {
            sync.Lock();

            // -1 means, that the pool should decide how many objects it
            // can create
            if (objectCount == -1) // calculate the number of objects
            {
                objectCount = Math.Min(maxSize - pool.Count, DefaultGrowSize);
            }
            if (objectCount <= 0) // pool has exhausted
            {
                sync.Unlock();
                return false;
            }
            for (int i = 0; i < objectCount; i++)
            {
                // create brand new objects
                ObjectType po = CreatePoolableObject();

                // activated at InternalDraw
                pool.Add(po, po);		// add to pool
                freeObjects.Push(po);	// ... and to the free objects list
            }
            sync.Unlock();
            return true;
        }

        #endregion

        #region Protected Members

        protected Type type;
        protected int minSize = PoolableAttribute.DefaultMinPoolSize;
        protected int maxSize = PoolableAttribute.DefaultMaxPoolSize;
        protected int preferedSize = PoolableAttribute.DefaultPreferedPoolSize;
        protected long creationTimeout = PoolableAttribute.DefaultCreationTimeout;
        protected static int DefaultGrowSize = 16;

        /// <summary>
        /// Obtains the type information.
        /// </summary>
        protected virtual void ObtainTypeInformation()
        {
            sync.Lock();

            // get the type's custom attributes
            object[] attArray = type.GetCustomAttributes(true);
            Attribute[] atts = (Attribute[])attArray;

            // search for our PoolableAttribute
            for (int i = 0; i < atts.Length; i++)
            {
                if (atts[i].GetType() == typeof(PoolableAttribute))
                {
                    // retrieve pooling information
                    PoolableAttribute pt = (PoolableAttribute)atts[i];
                    minSize = pt.MinPoolSize;
                    preferedSize = pt.PreferedPoolSize;
                    maxSize = pt.MaxPoolSize;
                    creationTimeout = pt.CreationTimeout;
                    break;
                }
            }
            sync.Unlock();
        }

        /// <summary>
        /// helper function, which creates poolable objects
        /// </summary>
        /// <returns></returns>
        protected virtual ObjectType CreatePoolableObject()
        {
            return Activator.CreateInstance<ObjectType>();
        }

        /// <summary>
        /// Determines whether [is free object to much].
        /// </summary>
        /// <returns>
        /// 	<c>true</c> if [is free object to much]; otherwise, <c>false</c>.
        /// </returns>
        protected virtual bool IsFreeObjectToMuch()
        {
            if (freeObjects.Count > pool.Count * 3 / 4 && pool.Count / 2 > minSize)
            {
                return true;
            }
            return false;
        }

        /// <summary>
        /// Shrinks the pool.
        /// </summary>
        protected virtual void ShrinkPool()
        {
            for (int i = 0; i < pool.Count / 2 && i < freeObjects.Count; i++)
            {
                pool.Remove(freeObjects.Pop());
            }
        }

        #endregion

        #region Constructors

        public ObjectPool() : this(null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:ObjectPool&lt;ObjectType&gt;"/> class.
        /// </summary>
        /// <param name="concreteType">Type of the concrete.</param>
        public ObjectPool(Type concreteType)
        {
            if (concreteType == null)
            {
                this.type = typeof(ObjectType);
            }
            else
            {
                this.type = concreteType;
            }

            this.canExhaust = true;
            this.created = false;
            this.pool = null;
            this.freeObjects = null;
            this.sync = new SyncGuard(this);

            // obtain the PoolableAttribute and its properties
            ObtainTypeInformation();
        }

        #endregion

        #region Public Members

        /// <summary>
        /// creates the actual pool
        /// </summary>
        /// <param name="configuration">The configuration.</param>
        /// <param name="canExhaust">if set to <c>true</c> [can exhaust].</param>
        /// <returns></returns>
	    public ObjectPool<ObjectType> CreatePool(bool canExhaust)
        {
		    if (created)
            {
			    throw new InvalidOperationException ("Pool has been already created");
            }
		    sync.Lock ();
		    this.canExhaust = canExhaust;

		    // create helper objects with capacity = the prefered size
		    pool = new Dictionary<object,ObjectType>(preferedSize);
		    freeObjects = new Stack<ObjectType>(preferedSize);
		    created = true;
		    sync.Unlock();

		    // create only "minimum pool size" objects
		    Grow(minSize); // create initial pool with minimum size
		    return this;
	    }

        /// <summary>
        /// Creates the pool.
        /// </summary>
        /// <returns></returns>
        public ObjectPool<ObjectType> CreatePool()
	    {
		    return CreatePool(true);
	    }

        /// <summary>
        /// Destroys the pool.
        /// </summary>
	    public void DestroyPool() 
	    {
		    if (!created)
            {
			    throw new InvalidOperationException("Pool cannot be destroyed, becuase it has not been created");
            }
		    sync.Lock();

		    // clear internal helper objects
		    pool.Clear();
		    freeObjects.Clear();
		    created = false;
		    sync.Unlock();
	    }

        /// <summary>
        /// Draws an object from the pool. If there aren't any objects created, a new one is created, otherwise an existing one is reused.
        /// </summary>
        /// <returns></returns>
	    public ObjectType Draw()
	    {
		    if (!created)
            {
			    throw new InvalidOperationException ("Pool has not been created");
            }

		    // this one is a poolable type
		    // search for available item in the pool and return it
		    ObjectType po = InternalDraw();
		    if (po != null)
		    {
			    return po;
		    }
		    else // no free ones
		    {
			    if (Grow (-1)) 
			    {
				    // after the grow there will be free ones now
				    return InternalDraw();
			    }
			    else
			    {
				    if (canExhaust)
				    {
					    throw new InvalidOperationException("Pool exhausted!");
				    }
				    else // the pool should wait until a free object returns
				    {
					    while (po == null)
					    {
						    po = InternalDraw();
						    Thread.Sleep(0); // give just a scrap of time
					    }
					    return po;
				    } // if (canExhaust)
			    } // if (Grow (-1))
		    } // if (po != 0)
	    } // Draw

        /// <summary>
        /// Returns an object to the pool to be reused
        /// </summary>
        /// <param name="obj">The obj.</param>
	    public void Return(ObjectType obj)
	    {
		    if (!created)
            {
			    throw new InvalidOperationException ("Pool has not been created");
            }
		    if (!pool.ContainsKey(obj))
            {
			    throw new InvalidOperationException ("Object was not created by this object pool");
            }

		    // cleanup and return the item in the pool
		    sync.Lock();
		    freeObjects.Push(obj);
		    sync.Unlock();
        }

        #endregion
    }
}
