using System;
using System.Collections.Generic;
using System.Text;

namespace Ilungasoft.Helper.ObjectPooling
{
    [AttributeUsageAttribute(AttributeTargets.Class, AllowMultiple = false)]
    public class PoolableAttribute : Attribute
    {
        #region Private Members

        private int minPoolSize;
	    private int preferedPoolSize;
	    private int maxPoolSize;
	    private long creationTimeout;

        /// <summary>
        /// Checks the state.
        /// </summary>
        private void CheckState()
        {
            if (minPoolSize > preferedPoolSize || minPoolSize > maxPoolSize || preferedPoolSize > maxPoolSize)
            {
                throw new ArgumentException("The condition min <= prefered <= max pool size was not met");
            }
            if (creationTimeout < -1)
            {
                creationTimeout = InfiniteCreationTimeout;
            }
        }

        #endregion

        #region Const Members

        public const int DefaultMinPoolSize = 1;
        public const int DefaultPreferedPoolSize = 16;
        public const int DefaultMaxPoolSize = 100000;
        public const long DefaultCreationTimeout = 60000;
        public const long MaxCreationTimeout = 0x80000000; // ~2 billions
	    public const long InfiniteCreationTimeout = -1;

        #endregion

        #region Public Members

        /// <summary>
        /// Initializes a new instance of the <see cref="T:PoolableAttribute"/> class.
        /// </summary>
        /// <param name="minPoolSize">Size of the min pool.</param>
        /// <param name="preferedPoolSize">Size of the prefered pool.</param>
        /// <param name="maxPoolSize">Size of the max pool.</param>
        /// <param name="creationTimeout">The creation timeout.</param>
        public PoolableAttribute(int minPoolSize, int preferedPoolSize, int maxPoolSize, long creationTimeout)
        {
            this.minPoolSize = minPoolSize;
            this.preferedPoolSize = preferedPoolSize;
            this.maxPoolSize = maxPoolSize;
            this.creationTimeout = creationTimeout;
            CheckState();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:PoolableAttribute"/> class.
        /// </summary>
        /// <param name="minPoolSize">Size of the min pool.</param>
        /// <param name="preferedPoolSize">Size of the prefered pool.</param>
        /// <param name="maxPoolSize">Size of the max pool.</param>
        public PoolableAttribute(int minPoolSize, int preferedPoolSize, int maxPoolSize) : this(minPoolSize, preferedPoolSize, maxPoolSize, DefaultCreationTimeout)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:PoolableAttribute"/> class.
        /// </summary>
        public PoolableAttribute()
            : this(DefaultMinPoolSize, DefaultPreferedPoolSize, DefaultMaxPoolSize, DefaultCreationTimeout)
        {
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets the size of the min pool.
        /// </summary>
        /// <value>The size of the min pool.</value>
        public int MinPoolSize
        {
            get
            {
                return minPoolSize;
            }
            set
            {
                minPoolSize = value;
                CheckState();
            }
        }

        /// <summary>
        /// Gets or sets the size of the max pool.
        /// </summary>
        /// <value>The size of the max pool.</value>
        public int MaxPoolSize
        {
            get
            {
                return maxPoolSize;
            }
            set
            {
                maxPoolSize = value;
                CheckState();
            }
        }

        /// <summary>
        /// Gets or sets the size of the prefered pool.
        /// </summary>
        /// <value>The size of the prefered pool.</value>
        public int PreferedPoolSize
        {
            get
            {
                return preferedPoolSize;
            }
            set
            {
                preferedPoolSize = value;
                CheckState();
            }
        }

        /// <summary>
        /// Gets or sets the creation timeout.
        /// </summary>
        /// <value>The creation timeout.</value>
        public long CreationTimeout
        {
            get
            {
                return creationTimeout;
            }
            set
            {
                creationTimeout = value;
                CheckState();
            }
        }

        #endregion
    }
}
